#! python3
# -*- coding: ISO-8859-1 -*-
#-------------------------------------------------------------------------------
# $Id$
#
# Copyright 2016 Jean-Luc PLOIX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#-------------------------------------------------------------------------------
#from driver._driverlib import filetype
'''Demo program for NN1
'''

import os
import sys
import stat
from six.moves.configparser import NoOptionError
from collections import defaultdict, OrderedDict
from random import randint
import time
from argparse import ArgumentParser, FileType
# from colorama import Fore, Style

from monal.util.toolbox import _linkstr
from nntoolbox.modelutils import getModelDict
from nntoolbox.timetb import second2str
from nntoolbox.modelutils import getGmxFileAndSave
from nntoolbox.filetoolbox import option2Arguments, writeArgumentFile, \
    arg_file_parse, adaptConfigStr2hidden, IndexFilename, decorfile, \
    getDictFromGmx
from nntoolbox.excelutils import createFileInfo, getBestTrainChart, \
    getTestChart, new_rows_line_chart, initiateExcelWriter, \
    finishExcel, getCustomDict, simpleRangeType, multiRangesType, \
    simpleRangeType, multiRangesType  #outputmarker,
from nntoolbox.utils import setEnvVar, splitvarunit, Namespace, finished, \
    hasOneOf, isInt, isIntBase, getListFromStr, strintlist, doNothing, \
    float_format, IDformat, roundoff_format, callerView

from monal import __version__ as monalversion  #, monalconst as C
from monal.util.monaltoolbox import yesNoQuestion, debugOutput, choiceQuestion
from monal.driver._driverlib import DriverLib
from monal.util.toolbox import toPrint, defaultedConfigParser
import nntoolbox.utils as nnu
from monal.datareader.modeldataframe import get_modelDataframe
from monal.util.utils import getuserbase, CCharOnlys
from monal.datareader.excelsource import getcustomproperty, getFrameFromExcel
from monal.util.lineartoolbox import gramSchmidt, cosTheta  #, randomProbe2,
from monal.modelutils import ReadModule

from nn1 import nn1_api_service as nn1serv, NN1_API as nn1api
from nntoolbox import __version__ as nntoolboxversion
from nn1 import __version__ as nn1version
from chem_gm import __version__ as gmversion

import warnings
warnings.filterwarnings("ignore", category=UserWarning)

debugrun = 0

def basicTrainTestAction(options, trainDataFrame, testDataFrame,
                         doprint, resultFrameList, excelactionlist,
                         mynames):
    """Train and test basic action.
    seed and hidden variables are fixed
    """

    excelwriter = resultFrame = trainFrame = testFrame = \
    looTrainFrame = looTestFrame = bsTrainFrame = bsTestFrame = None
    LOOResultTrainColumn = 0
    testresultColumn = -1
    startsloowithtest = True
    loowidth = 0
    traintype = ""
    starttimeloc = time.time()
    if hasattr(options, "configstrMem"):
        options.configstr = adaptConfigStr2hidden(options.configstrMem, options.hidden)
    trainfailed = False
    resfile = os.path.join(options.savedir, 'train.cfg')
    sqlFileName = ""
    sqlLOOFileName = ""
    gmxfile = ""
    gmxfileLOO = ""

    if 'train' in options.actionlist:
        doprint("running apiDemo Train")

        rest = nn1api.trainAction(options, sourceframe=trainDataFrame,
                                  doprint=doprint,
                                  resultFrameList=resultFrameList)
        libfilename, resultFrame, cfgfile, gmxfile, dico, sqlFileName = rest
        confResult = defaultedConfigParser(resfile)
        if not confResult.has_section(options.root):
            confResult.add_section(options.root)
        confResult.set(options.root, "cfgfile", cfgfile)
        confResult.save()

        trainfailed = resultFrame is None
        if trainfailed:
            doprint("Train failed")
        else:
            doprint("modelfile : {0}".format(callerView(gmxfile,
                                                        options.caller,
                                                        options.externpath)))
            doprint('apidemo Train done')
    if trainfailed:
        doprint('apidemo Failed')
        return
    
#     testindex = 0
#     negtestindex = 0
    maxrec = options.multimaxrec[0]
    options.maxrec = maxrec
    if options.copytype and hasOneOf(options.actionlist, excelactionlist):
        res = nn1serv.apiCreateExcelAndResultFrame(options, resultFrame,
                                                   excelwriter, resfile,
                                                   doprint)  # i ci et la  mynames.resultsheetname,
        excelwriter, resfile, resultFrame, tempres = res  # i ci et la
        confResult = defaultedConfigParser(resfile)
        countmax = min(options.maxrec, resultFrame.shape[0])
        root = options.root  #confResult.get("projects", "0")
        cfgfile = confResult.get(root, "cfgfile")
        config = defaultedConfigParser(cfgfile)
        paramcount = config.getint(root, "paramcount")
        
    nn1serv.saveFrameToExcel(options, resultFrame, excelwriter,
                             mynames.resultsheetname, 'result')        

    if options.notrain:
        if hasattr(options, "modelfile") and options.modelfile:
            gmxfile = options.modelfile
        else:
            gmxfile = config.get(root, 'gmxfile')
        try:
            mcount = config.getint(root, "mcount")
        except NoOptionError:
            mcount = -1
            
        pptyname = config.get(root, "propertyname")
        paramnames = config.get(root, "paramnames").split(";")
        outputnorm = config.get(root, "outputnorm").split(";")
        outputnorm = [float(val) for val in outputnorm]
        dataCount = config.getint(root, "baselen")
        options.seed = config.getint("training_%s"% pptyname, 'seed')
        if hasattr(options, "sqlFileName") and options.sqlFileName:
            sqlFileName = options.sqlFileName
        else:
            sqlFileName = config.get(root, "sqlfilename")
        gm0 = getModelDict(config, propertyName=pptyname, 
            parameterNames=paramnames, outputnorm=outputnorm, 
            paramCount=paramcount, dataCount=dataCount, 
            version=nn1version)
        
        if "ID" in resultFrame.columns:
            indexlist = resultFrame["ID"][:countmax]
        else:
            indexlist = resultFrame.index[:countmax]
        config.set(root, "mcount", "%d" % countmax)
        if not gmxfile or not os.path.exists(gmxfile):
            gmxfile, dico = getGmxFileAndSave(sqlFileName, gm0, 
                config, indexlist, options, full=True, 
                dosave=(mcount != countmax))
            config.set(root, "gmxfile", gmxfile)
        else:
            dico = getDictFromGmx(gmxfile)
    
    apiTrainRes = nn1serv.apiTrainResults(options, resfile, gmxfile=gmxfile, trainDataFrame=trainDataFrame) 
    trainFrame, IDlist, root, module, trainresultColumn, gmxfile = apiTrainRes
    
    nn1serv.saveFrameToExcel(options, trainFrame, excelwriter, mynames.trainsheetname, 'train')  #, wit hindex='identifiers' in options.grouplist)
    
    if 'test' in options.actionlist and options.test:
        #if not testindex:
        message0 = "launch test?" if options.notrain else  "continue and launch test?"
        message = "launch test" if options.verbose else ""
        if not yesNoQuestion(message0, 'y', message, forceyes=options.yes, doprint=doprint):
            doprint("aborted")
            finished(tempres)
            return
        doprint("running apiDemo Test")
        if options.notrain and (options.verbose >= 4):  # and not testindex: 
            doprint("working folder: %s"% (callerView(options.savedir, options.indirectoutput)))  #outputmarker
        cfgfile = cfgfile if cfgfile else options.cfgfile  
        datacount = trainFrame.shape[0]

        testFrame, testresultColumn = \
        nn1api.testAction(options, testDataFrame, datacount, IDlist, 
            cfgfile, gmxfile, dico, doprint=doprint)
            
        doprint('apidemo Test done')
    
        if options.verbose >= 5:  
            formatters = {'RoundoffTest': roundoff_format(), 'ID': IDformat()}
            doprint(testFrame.to_string(float_format=float_format(options.verbose), formatters=formatters))
            
        nn1serv.saveFrameToExcel(options, testFrame, excelwriter, mynames.testsheetname, 'test')  #, withi ndex='identifiers' in options.grouplist)
    
    if 'loo'in options.actionlist: 
        message0 = "launch Leave-One-Out?" if options.notrain else  "continue and launch Leave-On?"
        message = "launch Leave-One_out" if options.verbose else ""
        if not yesNoQuestion(message0, 'y', message, forceyes=options.yes, doprint=doprint):
            doprint("aborted")
            finished(tempres)
            return
        doprint("Running leave-One-Out train")          

        if module is None:
            module = nn1serv.LoadModuleFromCfg(config)
        elif config.has_section(root):
            gmxfile = config.get(root, "gmxfile")
            dico = getDictFromGmx(gmxfile)
            module.loadParameters(dico)
        if trainDataFrame is not None:
            datacount, datasize = module.loadTrainingDataFromFrame(trainDataFrame)
        
#        looTrainFrame, lootrainfailed, LOOResultTrainColumn, gmxfileLOO = \
        res = nn1api.looTrainAction(options, module, trainDataFrame, 
            resultFrame, doprint=doprint) 
        looTrainFrame, lootrainfailed, LOOResultTrainColumn, gmxfileLOO, sqlLOOFileName = res
        nn1serv.saveFrameToExcel(options, looTrainFrame, excelwriter, mynames.lootrainsheetname, 'lootrain')  #, with index='identifiers' in options.grouplist)
        doprint("LOOmodelfile : {0}".format(callerView(gmxfileLOO, options.caller, options.externpath)))
        doprint("Leave-One-Out train done")          
        startsloowithtest = (looTrainFrame is None) and not lootrainfailed  
        
    
    if 'lootest' in options.actionlist: 
        message0 = "launch Leave-One-Out test?" if options.notrain else  "continue and launch Leave-One-Out test?"
        message = "launch Leave-One-Out test" if options.verbose else ""
        if not yesNoQuestion(message0, 'y', message, forceyes=options.yes, doprint=doprint):
            finished(tempres)
            doprint("aborted")
            finished(tempres)
            return
        doprint("Running leave-One-Out test")          
        if startsloowithtest and (options.verbose >= 4):  # and not testindex:
            doprint("working folder: %s\n"% (callerView(options.savedir, options.indirectoutput)))  #outputmarker
        if  startsloowithtest:
            lootrainsheetname = mynames.lootrainsheetname
            looTrainFrame = getFrameFromExcel(options.origin, 'LOO training results', reduce=True)
            LOOResultTrainColumn = testDataFrame.shape[1]
        else:
            lootrainsheetname = ""   


        if apiTrainRes is None:
            apiTrainRes = nn1api.apiTrainResults(options, resfile, extratype="loo") 
        looTrainFrame, IDlist, root, module, _, _ = apiTrainRes
        if startsloowithtest:
            nn1serv.saveFrameToExcel(options, looTrainFrame, excelwriter, mynames.lootrainsheetname, 'loo')  #, with index='identifiers' in options.grouplist)
#             looTrainFrame, IDlist, root, module, _, _ = apiTrainRes
#             if (looTrainFrame is not None) and options.copytype:
#                 if options.copyexcel:
#                     looTrainFrame.to_excel(excelwriter, lootrainsheetname, index=True)  
#                 elif options.copytxtcsv and os.path.exists(options. targetfolder):
#                     fname = os.path.join(options. targetfolder, 'lootrainFrame')
#                     fname = decorfile(fname, options, ".%s"% options.copytype, indextemplate="(%d)")
#                     with open(fname, 'w') as ff:
#                         ff.write(looTrainFrame.to_csv(sep=options.sep, decimal=options.decimal))

            
        looTestFrame = nn1api.looTestAction(options, testDataFrame, resfile, IDlist, doprint=doprint) #tempres, excelwriter, lootrainsheetname, apiTrainRes, 
                  
        nn1serv.saveFrameToExcel(options, looTestFrame, excelwriter, mynames.lootestsheetname, 'lootest')  #, with index='identifiers' in options.grouplist)

        doprint("Leave-One-Out test done")          
    
    if 'bootstrap' in options.actionlist:
        if module is None:
            module = nn1serv.LoafModuleFromCfg(config)
            datacount, datasize = module.loadTrainingDataFromFrame(trainDataFrame)
        nn1api.bsTrainAction(options, module, doprint=doprint)
    
    
    if 'bootstraptest' in options.actionlist:
        nn1api.bsTestAction(options, doprint=doprint)
    
    
    if 'usage' in options.actionlist:
        nn1api.usageAction(options, options.origin, options.data)

    
    if options.dotimer and (len(options.multihidden) > 1): 
        delta = float(time.time() - starttimeloc)
        doprint("running time for {0} hidden neurons: {1}".format(options.hidden, second2str(delta)))

    
    libfilename = options.modulename
    if not libfilename and cfgfile:
        config = defaultedConfigParser(cfgfile)
        libfilename = config.get(root, "module")
    if not os.path.exists(libfilename):
        libfilename = os.path.join(options.savedir, libfilename)
    excelfilename = nn1serv.getExcelFileName(options)
    dico = OrderedDict({
        'nn1 version' : ( nn1version, 'text'),
        'chem_gm version': (gmversion, 'text'),
        'monal version': (monalversion, 'text'),
        'nntoolbox version': (nntoolboxversion, 'text'),
        'Working folder': (options.savefolder, 'text'),
        'Result file': (excelfilename, 'text'),
    })
    if sqlFileName:
        dico["SQL file"] = (sqlFileName, 'text')
    if sqlLOOFileName:
        dico["SQL LOO file"] = (sqlLOOFileName, 'text')
    if gmxfile:
        dico["Model file"] = (gmxfile, 'text')
    if gmxfileLOO:
        dico["Model LOO file"] = (gmxfileLOO, 'text')
        
    gmCustomProperties = getCustomDict(options, dico)

    if options.copytype and libfilename: 
        sheetList = [mynames.resultsheetname, mynames.trainsheetname, mynames.testsheetname, mynames.lootrainsheetname, mynames.lootestsheetname, mynames.bstrainsheetname, mynames.bstestsheetname] 
        frameList = [resultFrame, trainFrame, testFrame, looTrainFrame, looTestFrame, bsTrainFrame, bsTestFrame]
        modeldic, _ = ReadModule(libfilename, pptyname=options.pptyname, dodelete=True)
        modeldic['bestCount'] = sorted(options.multimaxrec)
        excelresult = finishExcel(excelwriter,
            options, 
            gmCustomProperties, 
            sheetList, 
            frameList, 
            modeldic, 
            libfilename, 
            IDlist, 
            trainresultColumn,  
            testresultColumn, 
            LOOResultTrainColumn,
            loowidth, 
            traintype, 
            tempres)
        if excelresult and cfgfile:
            config = defaultedConfigParser(cfgfile)
            #st = _linkstr("training", options.pptyname) if options.pptyname else "training"
            config.set('training', 'excelresult', excelresult)
            config.save()
    else:
        doprint("no copy")

    return modeldic, gmCustomProperties
    
def manageApiOptionsResult(options, toprint, stopcode, doprint, encoding):
    """Processing special asking (stop, display) in options.
    Return True to stop the program.
    """
    if stopcode & nnu.OPTION_STOP:
        doprint(toprint, file=sys.stderr)
        return 1      
    
    if stopcode & nnu.OPTION_DISPLAY_FILE_AND_STOP:
        # affichage la licence ou de l'aide prise dans le fichier "toprint"
        #if (ou tfile is not None):
        if int(options.verbose) >= 5:
            doprint(toprint, file=sys.stderr)
#            doprint()
            doprint(encoding)
            doprint()
        with open(toprint, 'rb') as ff:  #, encoding='UTF-8'
            line = ff.readline()
            while line:
                #val = line.decode("ISO-8859-15") 
                val = line.decode("UTF-8") 
                doprint(val, file=sys.stderr) 
                line = ff.readline()
            # val = ff.read().decode("ISO-8859-15") 
        doprint()
        return 2
    
    elif stopcode & nnu.OPTION_WARN_AND_STOP: 
        # affichage du message toprint   
        doprint(toprint, file=sys.stderr)
        return 3
    
    if options.verbose >= 5:
        import platform
        doprint("Python", platform.python_version())               
    
    if toprint:
        # affichage du message
        doprint(toprint)
        if stopcode & nnu.OPTION_WARN_AND_ASK: 
            # test si possibilite d'arret
            message = "corrected options accepted"
            if not yesNoQuestion("continue anyway?", 'y', message, forceyes=options.yes, doprint=doprint):
                doprint("aborted")
                finished("")
                return 4
    return 0
    
def runOnce(argv=[], externaloptions=None):
    if debugrun:
        for ind, val in enumerate(argv):
            print("argv[{0}] : {1}".format(ind, val))
    
    starttime = time.time()
    excelactionlist = ['train', 'test', 'loo', 'lootest', 'bootstrap', 'bootstraptest']   
    namesdict = {
        "libfilename" : "",
        "resultsheetname" :'Training summary',
        "testsheetname" : 'Test results',
        "tstshname" : "",
        "lootestsheetname" : 'LOO test results',
        "lootstshname" : "",
        "bstestsheetname" : 'Bootstrap test results',
        "bsstshname" : "",
        "trainsheetname" : 'Training results',
        "trnshname" : "",
        "lootrainsheetname" : 'LOO training results',
        "bstrainsheetname" : 'Bootstrap training results',
      }
    mynames = Namespace(**namesdict)
    
    tempres = ""
    toprint = ""
    stopcode = False

    options, toprint, stopcode = nn1api.api_options(argv[1:], proposedOptions=externaloptions)

    try:
        options.debug |= nn1serv.demodebug
    except: pass
    outfile = options.outfile
    doprint = toPrint(outfile) 
    if outfile:
        encoding = outfile.encoding  
        
    if manageApiOptionsResult(options, toprint, stopcode, doprint, encoding):
        # il y a eu une erreur ou une demande d'affichage, On sort du programme.
        return                 
    
    if 'usage' in options.action:
        # demande d'usage
        return nn1api.usageAction(options)
    
    if not options.source and 'train' in options.actionlist:
        raise nn1serv.monalError("This program needs a data file.")
    elif not options.source and not options.test and 'test' in  options.actionlist:
        raise nn1serv.monalError("This program needs a test file.") 
    
    # sourceAnalysis est toujours appele
    
    createArgumentFile = 'analyse' in options.actionlist or ('preproc' in options.actionlist and len(options.actionlist)==1)
    
        # Lecture et analyse d'un nouveau fichier de donnes, creation du fichier d'arguments
    options = nn1api.sourceAnalysis(options, doprint, forceAnalyse=createArgumentFile) #
    if options is None:
        doprint("Input error. Program stopped !!!")
    elif nn1api.preProcAction(options, doprint):
        doprint("PreProc action error. Program stopped !!!")
        options = None
    
    if options is None:
        return -1   
    
    if createArgumentFile:
        if options.notest:
            forceAction = "train"
        else:
            forceAction = "train,test"
        return nn1api.saveOptionsToFile(options, forceAction=forceAction, forceYes=True, forceModel=True, doprint=doprint)  #forceAction="train,test", rootTargetName="arguments"

#######################################################################@@
    trainDataFrame = None
    filetype = options.filetype
    grouplist = options.grouplist
    if nn1api._hasTrain(options):
        datafile = options.source
        datarange = options.datarange
        fieldlist = options.datafields
        drop = ('identifiers' in grouplist) and not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
        trainDataFrame = get_modelDataframe(datafile, datarange, fieldlist, grouplist, filetype=filetype, drop=drop, withindexdict=False) 

    testDataFrame = None
    if nn1api._hasTest(options):
        if not options.test:
            options.test = options.source
        if not options.testrange:
            options.testrange = 'TEST'
        if not options.testfields:
            options.testfields = options.datafields
        datafile = options.test
        datarange = options.testrange
        fieldlist = options.testfields
        drop = ('identifiers' in grouplist) and not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
        try:
            testDataFrame = get_modelDataframe(datafile, datarange, fieldlist, grouplist, filetype=filetype, drop=drop, withindexdict=False) 
        except AttributeError:
            if drop:                
                fieldlist = fieldlist[1:]
                newfieldlist = [[val - 1 for val in lst] for lst in fieldlist]
                grouplist = grouplist[1:]
                drop = ('identifiers' in grouplist) and not grouplist.index('identifiers') and (len(fieldlist[0]) == 1)
                testDataFrame = get_modelDataframe(datafile, datarange, newfieldlist, grouplist, filetype=filetype, drop=drop, withindexdict=False)
            else:
                testDataFrame = None
            
        
    if not 'usage' in options.action:
        if not trainDataFrame:
            options.propertyName = testDataFrame.fieldsOf('outputs')[0]
            options.propertyunit = testDataFrame.units[options.propertyName]
            options.inputcount = len(testDataFrame.fieldsOf('inputs'))
        # 
            origin = nn1serv.getOrigin(options)
            if origin and os.path.isfile(origin):
                filename = origin 
                
        else:
            options.propertyName = trainDataFrame.fieldsOf('outputs')[0]
            options.propertyunit = trainDataFrame.units[options.propertyName]
            options.inputcount = len(trainDataFrame.fieldsOf('inputs'))
    
    options.tableName = 'trainingResult'
    options.inoutcount = 0
    memtarget = options.targetfolder[:]
    for seed in options.multiseed:
        # boucle sur les graines aleatoires (seed)
        excelresult = ""
        excelwriter = None
        cfgfile = ""
        resfile = ""
        trainfailed = False
        dico = None
        resultFrameList = []
        if seed < 0:
            seed = randint(0x10, 0xFFFFFFFF)    
        options.seed = seed
        if len(options.multihidden) > 1:
            # si on est en multihidden, on cree un repertoire ici pour la seed en cours
            st = strintlist(options.multihidden, delimiter="-", 
                groupdelimiter="--", compact=True)
            dirname = "{0}-SEED{1}-{2}N".format(options.root, seed, st)
            options.targetfolder = os.path.join(memtarget, dirname)
            options.targetfolder = IndexFilename(options.targetfolder)
            if not os.path.exists(options.targetfolder):
                os.makedirs(options.targetfolder)
            if options.argumentfile:
                targetfile = os.path.join(options.targetfolder, "arguments.txt")
                with open(options.argumentfile, 'r') as sf:
                    with open(targetfile, 'w') as tf:
                        for line in sf:
                            if len(line) and not line.startswith("#"):
                                lst = line.split("=")
                                if (len(lst) > 1) and lst[0].strip() == "SEED":
                                    lst[1] = str(seed)
                                    if line.endswith("\n"):
                                        lst[1] += "\n"  
                                line = "=".join(lst)
                            tf.write(line)

        for options.hidden in options.multihidden:  # 
            # boucle sur les neurones caches (hidden)
            modeldic, gmCustomProperties = basicTrainTestAction(options, trainDataFrame, testDataFrame, doprint, resultFrameList, excelactionlist, mynames)
        # fin boucle sur hidden

        if len(options.multihidden) > 1:
            nn1serv.excelRecapitulation(options, resultFrameList, modeldic, gmCustomProperties, doprint)
                
    # fin boucle sur seed

    if options.dotimer: 
        delta = float(time.time() - starttime)
        doprint("running time {0}".format(second2str(delta)))
    finished(tempres)                
           
def run(argv=["_"]):
    res = runOnce(argv)
    if isinstance(res, tuple) and len(res) >= 2:
        options = res[0]
        argname = res[1]
        if argname and os.path.exists(argname):
            mess = "do you want to run the new project ?"
            if yesNoQuestion(mess, default="y", outfile=sys.stdout, 
                forceyes=options.yesyes):  #, verbose=verbose)
                argumentfile = "@{0}".format(argname)
                res = runOnce(["", argumentfile], options)
    return res
    
if __name__ == '__main__':
    run(sys.argv)
    print("Done")